#!/usr/bin/env python3
"""
 @file examples/schur_obj.py

 PaStiX Schur python example with an object oriented programing solution.

 @copyright 2017-2024 Bordeaux INP, CNRS (LaBRI UMR 5800), Inria,
                      Univ. Bordeaux. All rights reserved.

 @version 6.4.0
 @author Pierre Ramet
 @author Mathieu Faverge
 @author Louis Poirel
 @date 2024-07-05

 This example shows how to use pastix solver to solve a system with a
 Schur complement.

"""
##\cond
import pypastix as pastix
import scipy.sparse as sps
import scipy.linalg as la
import numpy as np

# Hack to make sure that the mkl is loaded
tmp = np.eye(2).dot(np.ones(2))

# Set matrix A from Scipy Sparse matrices
n = 30
A = sps.spdiags([np.ones(n)*i for i in [4., -1, -1]],
                [0, 1, -1], n, n)

x0 = np.arange(n).reshape(n,1)
# Construct b as b = A * x_0
b = A.dot(x0)
x = b.copy()

# Initialize the Schur list as the first third of the elements
nschur = min( int(n / 3), 5000 )
schurlist = np.arange(nschur)

solver = pastix.solver(verbose=2)

solver.schur(A, schurlist)
S = solver.S
f = solver.schur_forward(x)
y = la.solve(S, f, sym_pos=True, lower=True)
x = solver.schur_backward(y)

# Check the solution
rc = solver.check( x, b, x0=x0 )

solver.finalize()

exit(rc)
##\endcond



"""
 @file spm.py

 SPM python wrapper

 @copyright 2017-2024 Bordeaux INP, CNRS (LaBRI UMR 5800), Inria,
                      Univ. Bordeaux. All rights reserved.

 @version 1.2.4
 @author Pierre Ramet
 @author Mathieu Faverge
 @author Tony Delarue
 @author Valentine Bregeot
 @date 2024-06-27

 @ingroup wrap_python

"""
from ctypes import *
import numpy as np

__use_sps__ = True
try:
    import scipy.sparse as sps
except ImportError:
    __use_sps__ = False

from .enum    import *
from .enum    import __spm_int__
from .enum    import __spm_mpi_enabled__
from .__spm__ import *

mpi_enabled = __spm_mpi_enabled__

class spmatrix():

    dtype = None

    def __init__( self, A=None, mtxtype_=mtxtype.General, driver=None, filename="", comm=pyspm_default_comm ):
        """
        Initialize the SPM wrapper by loading the libraries
        """
        if mtxtype_ == mtxtype.SymPosDef:
            mtxtype_ = mtxtype.Symmetric
        if mtxtype_ == mtxtype.HerPosDef:
            mtxtype_ = mtxtype.Hermitian

        self.spm_c = pyspm_spmatrix_t( mtxtype_,        #Matrix type
                                       coeftype.Double, #Floating point airthmetic
                                       fmttype.CSC,     #Format
                                       0,               #Baseval
                                       0,               #gN
                                       0,               #n
                                       0,               #gnnz
                                       0,               #nnz
                                       0,               #gNexp
                                       0,               #nexp
                                       0,               #gnnzexp
                                       0,               #nnzexp
                                       1,               #dof
                                       None,            #dofs
                                       layout.ColMajor, #layout dof
                                       None,            #colptr
                                       None,            #rowptr
                                       None,            #loc2glob
                                       None,            #values
                                       None,            #glob2loc
                                       0,               #clustnum
                                       1,               #clustnbr
                                       pyspm_convert_comm( comm ),  #comm
                                       1 ) # replicated
        self.id_ptr = pointer( self.spm_c )
        self.init( comm )

        if A is not None:
            self.fromsps( A, mtxtype_ )
        elif driver is not None:
            self.fromdriver( driver, filename )

    if __use_sps__:
        def fromsps( self, A, mtxtype_=mtxtype.General ):
            """
            Initialize the SPM wrapper by loading the libraries
            """

            if not sps.isspmatrix(A):
                raise TypeError( "A must be of type scipy.sparse matrix" )

            # Assume A is already in Scipy sparse format
            self.dtype = A.dtype
            flttype = coeftype.getptype( A.dtype )
            if flttype == -1:
                raise TypeError( "Invalid data type. Must be part of (f4, f8, c8 or c16)" )

            A = sps.csc_matrix( A )
            A.sort_indices()

            # Pointer variables
            self.py_colptr = np.array( A.indptr[:],  dtype=__spm_int__ )
            self.py_rowptr = np.array( A.indices[:], dtype=__spm_int__ )
            self.py_values = np.array( A.data[:] )

            if mtxtype_ == mtxtype.SymPosDef:
                mtxtype_ = mtxtype.Symmetric
            if mtxtype_ == mtxtype.HerPosDef:
                mtxtype_ = mtxtype.Hermitian

            self.spm_c.replicated = 1
            self.spm_c.baseval  = 0
            self.spm_c.mtxtype  = mtxtype_
            self.spm_c.flttype  = flttype
            self.spm_c.fmttype  = fmttype.CSC
            self.spm_c.n        = A.shape[0]
            self.spm_c.nnz      = A.getnnz()
            self.spm_c.dof      = 1
            self.spm_c.dofs     = None
            self.spm_c.layout   = layout.ColMajor
            self.spm_c.colptr   = self.py_colptr.ctypes.data_as(POINTER(__spm_int__))
            self.spm_c.rowptr   = self.py_rowptr.ctypes.data_as(POINTER(__spm_int__))
            self.spm_c.loc2glob = None
            self.spm_c.values   = self.py_values.ctypes.data_as(c_void_p)

            self.id_ptr = pointer( self.spm_c )

            self.updateComputedField()
            self.checkAndCorrect()

        def tosps( self ):
            """
            Return a Scipy sparse matrix
            """
            n      = int( self.spm_c.n )
            nnz    = int( self.spm_c.nnz )
            cflt   = coeftype.getctype(  self.spm_c.flttype )
            nflt   = coeftype.getnptype( self.spm_c.flttype )
            colptr = self.py_colptr.copy()
            rowptr = self.py_rowptr.copy()
            values = self.py_values.copy()

            baseval = colptr[0]
            colptr = colptr - baseval
            rowptr = rowptr - baseval

            return sps.csc_matrix((values, rowptr, colptr), shape=(n, n))

    def fromdriver( self, driver=driver.Laplacian, filename="10:10:10" ):
        """
        Initialize the SPM wrapper by loading the libraries
        """
        if filename == "":
            raise ValueError("filename must be prodived")

        pyspm_spmReadDriver( driver, filename.encode('utf-8'), self.id_ptr )

        self.dtype = coeftype.getnptype( self.spm_c.flttype )
        self.checkAndCorrect()

    def scale( self, alpha ):
        return pyspm_spmScal( float(alpha), self.id_ptr )

    def norm( self, ntype=normtype.Frobenius ):
        return pyspm_spmNorm( ntype, self.id_ptr )

    def base( self, baseval ):
        pyspm_spmBase( self.id_ptr, baseval )

    def findBase( self ):
        return pyspm_spmFindBase( self.id_ptr )

    def updateComputedField( self ):
        pyspm_spmUpdateComputedFields( self.id_ptr )

    def printInfo( self ):
        pyspm_spmPrintInfo( self.id_ptr )

    def printSpm( self ):
        pyspm_spmPrint( self.id_ptr )

    def init( self, comm=pyspm_default_comm ):
        pyspm_spmInitDist( self.id_ptr, comm )

    def checkAndCorrect( self ):
        spm1 = self.id_ptr
        spm2 = spmatrix()
        rc = pyspm_spmCheckAndCorrect( self.id_ptr, spm2.id_ptr )
        if ( rc == 0 ):
            return
        self = spm2
        self.spm_c = cast( spm2, POINTER(pyspm_spmatrix_t) ).contents
        self.id_ptr = pointer( self.spm_c )
        self.py_colptr = np.frombuffer( (__spm_int__ * (n+1)).from_address( cast(self.spm_c.colptr, c_void_p).value ), __spm_int__ ).copy()
        self.py_rowptr = np.frombuffer( (__spm_int__ *  nnz ).from_address( cast(self.spm_c.rowptr, c_void_p).value ), __spm_int__ ).copy()
        self.py_values = np.frombuffer( (cflt        *  nnz ).from_address( self.spm_c.values ), nflt ).copy()

    def __checkVector( self, n, nrhs, x ):
        if x.dtype != self.dtype:
            raise TypeError( "Vectors must use the same arithmetic as the spm" )

        if x.ndim > 2:
            raise TypeError( "Vectors must be of dimension 1 or 2" )

        if x.shape[0] < n:
            raise TypeError( "Vectors must be of size at least ", n )

        if (x.ndim == 1 and nrhs > 1) or (x.ndim>1 and x.shape[1] < nrhs):
            raise TypeError( "At least nrhs vectors must be stored in the vector" )

    def checkAxb( self, x0, b, x, nrhs=-1, **kwargs ):
        # if libspm == None:
        #     raise EnvironmentError( "SPM Instance badly instanciated" )

        eps = kwargs.setdefault( 'eps', -1. )

        b = np.array(b, self.dtype)
        x = np.asarray(x, self.dtype)

        n = self.spm_c.n
        if nrhs == -1:
            if x.ndim == 1:
                nrhs = 1
            else:
                nrhs = x.shape[1]

        if x0 is not None:
            x0 = np.asarray(x0, self.dtype)
            self.__checkVector( n, nrhs, x0 )
            ldx0 = x0.shape[0]
            x0ptr = x0.ctypes.data_as(c_void_p)
        else:
            ldx0 = 1
            x0ptr = None
        self.__checkVector( n, nrhs, b )
        self.__checkVector( n, nrhs, x )

        ldb  = b.shape[0]
        ldx  = x.shape[0]

        return pyspm_spmCheckAxb( eps, nrhs, self.id_ptr,
                                  x0ptr, ldx0,
                                  b.ctypes.data_as(c_void_p), ldb,
                                  x.ctypes.data_as(c_void_p), ldx )

    def genRHS( self, rhstype=rhstype.One, nrhs=1, getx=False ):
        # if libspm == None:
        #     raise EnvironmentError( "SPM Instance badly instanciated" )

        n = self.spm_c.n
        if nrhs == 1:
            b = np.zeros((n), self.dtype )
        else:
            b = np.zeros((n, nrhs), self.dtype)
        ldb = b.shape[0]
        self.__checkVector( n, nrhs, b )

        if getx:
            if nrhs == 1:
                x = np.zeros((n), self.dtype)
            else:
                x = np.zeros((n, nrhs), self.dtype)
            ldx  = x.shape[0]
            self.__checkVector( n, nrhs, x )
            xptr = x.ctypes.data_as( c_void_p )
        else:
            x    = None
            ldx  = 1
            xptr = None

        info = pyspm_spmGenRHS( rhstype, nrhs, self.id_ptr,
                                xptr, ldx, b.ctypes.data_as( c_void_p ), ldb )
        return x, b

    def mult( self, B, C, trans=trans.NoTrans, n=-1, alpha=1.0, beta=0. ):

        m = self.spm_c.n

        B = np.asarray(B, self.dtype)
        C = np.asarray(C, self.dtype)

        if n == -1:
            if B.ndim == 1:
                n = 1
            else:
                n = B.shape[1]

        self.__checkVector( m, n, B )
        self.__checkVector( m, n, C )

        ldb  = B.shape[0]
        ldc  = B.shape[0]

        if n == 1:
            pyspm_spmMatVec( trans, alpha, self.id_ptr,
                             B.ctypes.data_as(c_void_p),
                             beta,
                             C.ctypes.data_as(c_void_p) )
        else:
            pyspm_spmMatMat( trans, n, alpha, self.id_ptr,
                             B.ctypes.data_as(c_void_p), ldb,
                             beta,
                             C.ctypes.data_as(c_void_p), ldc )

    def save( self, filename="./matrix.spm" ):
        return pyspm_spmSave( self.id_ptr, filename )

    def load( self, filename="./matrix.spm" ):
        return pyspm_spmLoad( self.id_ptr, filename )



#!/usr/bin/env python3
"""
 @file spm_driver.py

 @brief SpM example to generate a sparse matrix from the spm drivers

 @copyright 2017-2024 Bordeaux INP, CNRS (LaBRI UMR 5800), Inria,
                      Univ. Bordeaux. All rights reserved.

 @version 1.2.4
 @author Pierre Ramet
 @author Mathieu Faverge
 @author Tony Delarue
 @author Alycia Lisito
 @date 2024-05-29

 @ingroup examples_python
 @code

 @endcode
"""

##\cond
import spm
import numpy as np

# Hack to make sure that the mkl is loaded
tmp = np.eye(2).dot(np.ones(2))

# Load a sparse matrix from the Laplacian driver
A = spm.spmatrix( driver=spm.driver.Laplacian, filename="10:10:10:2.:1." )

# Example from a HB file
#A = spm( driver=driver.HB, filename="$SPM_DIR/test/matrix/orsirr.rua" )

A.printInfo()

# Scale A for low-rank: A / ||A||_f
norm = A.norm()
A.scale( 1. / norm )

# Generate b and x0 vectors such that A * x0 = b
nrhs = 10
x0, b = A.genRHS( spm.rhstype.RndX, nrhs, True )

# Check that A * x = b
A.checkAxb( None, b, x0 )
##\endcond



#!/usr/bin/env python3
"""
 @file spm_scipy.py

 @brief SpM example to generate a sparse matrix from Scipy to SPM

 @copyright 2017-2024 Bordeaux INP, CNRS (LaBRI UMR 5800), Inria,
                      Univ. Bordeaux. All rights reserved.

 @version 1.2.4
 @author Pierre Ramet
 @author Mathieu Faverge
 @author Tony Delarue
 @author Alycia Lisito
 @date 2024-05-29

 @ingroup examples_python
 @code

 @endcode
"""

##\cond
import spm
import scipy.sparse as sps
import numpy as np

# Hack to make sure that the mkl is loaded
tmp = np.eye(2).dot(np.ones(2))

# Set the problem
n = 9
A = sps.spdiags([np.ones(n)*i for i in [4, -1, -1, -1, -1]],
                [0, 1, 3, -1, -3], n, n)
x = np.arange(n)
b = np.zeros(n)

spmA = spm.spmatrix( A )

# b <= A * x
spmA.mult( x, b, trans=spm.trans.NoTrans, alpha=1., beta=0. )

# Use the solver check to check that A * x == b
spmA.checkAxb( None, b, x )
##\endcond



"""
PyPaStiX
========

 @file __init__.py

 PaStiX python module intialization

 @copyright 2017-2024 Bordeaux INP, CNRS (LaBRI UMR 5800), Inria,
                      Univ. Bordeaux. All rights reserved.

 @version 6.4.0
 @author Pierre Ramet
 @author Mathieu Faverge
 @author Louis Poirel
 @date 2024-07-05

"""
import ctypes
import ctypes.util
import spm

# Load the PASTIX library
libpastix_name = ctypes.util.find_library('pastix')
if libpastix_name == None:
    raise EnvironmentError("Could not find shared library: pastix."
                           "The path to libpastix.so should be in "
                           "$LIBRARY_PATH")
try:
    libpastix = ctypes.cdll.LoadLibrary(libpastix_name)
except:
    raise EnvironmentError("Could not load shared library: pastix."
                           "The path to libpastix.so should be in "
                           "$LD_LIBRARY_PATH or $DYLD_LIBRARY_PATH on MacOS");

__all__ = [ 'libpastix' ]

from .enum   import *
from .pastix import *
from .solver import *



#!/usr/bin/env python3
"""
 @file examples/simple.py

 PaStiX simple python example

 @copyright 2017-2024 Bordeaux INP, CNRS (LaBRI UMR 5800), Inria,
                      Univ. Bordeaux. All rights reserved.

 @version 6.4.0
 @author Pierre Ramet
 @author Mathieu Faverge
 @author Louis Poirel
 @date 2024-07-05

"""
##\cond
import spm
import pypastix as pastix
import scipy.sparse as sps
import numpy as np

# Hack to make sure that the mkl is loaded
tmp = np.eye(2).dot(np.ones(2))

# Load a sparse matrix from HB driver
spmA = spm.spmatrix( None, driver=spm.driver.Laplacian, filename="10:10:10:2.:1." )
#spmA = spm.spmatrix( None, driver=spm.driver.HB, filename="$PASTIX_DIR/test/matrix/orsirr.rua" )
spmA.printInfo()

# Scale A for low-rank: A / ||A||_f
norm = spmA.norm()
spmA.scale( 1. / norm )

# Generate b and x0 vector such that A * x0 = b
nrhs = 1
x0, b = spmA.genRHS( spm.rhstype.RndX, nrhs )

# Initialize parameters to default values
iparm, dparm = pastix.initParam()

# Startup PaStiX
pastix_data = pastix.init( iparm, dparm )

# Change some parameters
iparm[pastix.iparm.factorization] = pastix.factotype.LU

# Perform analyze
pastix.task_analyze( pastix_data, spmA )

# Perform numerical factorization
pastix.task_numfact( pastix_data, spmA )

# Perform solve
x = b.copy()
pastix.task_solve( pastix_data, spmA, x )

# Refine the solution
pastix.task_refine( pastix_data, spmA, b, x )

# Check solution
rc = spmA.checkAxb( x0, b, x )

pastix.finalize( pastix_data, iparm, dparm )

exit(rc)
##\endcond



"""

 @file __spm__.py

 SPM python wrapper

 @copyright 2017-2024 Bordeaux INP, CNRS (LaBRI UMR 5800), Inria,
                      Univ. Bordeaux. All rights reserved.

 @version 1.2.4
 @author Pierre Ramet
 @author Mathieu Faverge
 @author Tony Delarue
 @date 2024-06-27

 This file has been automatically generated with gen_wrappers.py

 @ingroup wrap_python

"""
from ctypes import *
import numpy as np

from . import libspm
from .enum import __spm_int__
from .enum import __spm_mpi_enabled__

if __spm_mpi_enabled__:
    from mpi4py import MPI
    if MPI._sizeof(MPI.Comm) == sizeof(c_long):
        pyspm_mpi_comm = c_long
    elif MPI._sizeof(MPI.Comm) == sizeof(c_int):
        pyspm_mpi_comm = c_int
    else:
        pyspm_mpi_comm = c_void_p

    pyspm_default_comm = MPI.COMM_WORLD

    def pyspm_convert_comm( comm ):
        comm_ptr = MPI._addressof(comm)
        return pyspm_mpi_comm.from_address(comm_ptr)
else:
    pyspm_mpi_comm = c_int

    pyspm_default_comm = 0

    def pyspm_convert_comm( comm ):
        return c_int(comm)

class pyspm_spmatrix_t(Structure):
    _fields_ = [("mtxtype",     c_int               ),
                ("flttype",     c_int               ),
                ("fmttype",     c_int               ),
                ("baseval",     __spm_int__         ),
                ("gN",          __spm_int__         ),
                ("n",           __spm_int__         ),
                ("gnnz",        __spm_int__         ),
                ("nnz",         __spm_int__         ),
                ("gNexp",       __spm_int__         ),
                ("nexp",        __spm_int__         ),
                ("gnnzexp",     __spm_int__         ),
                ("nnzexp",      __spm_int__         ),
                ("dof",         __spm_int__         ),
                ("dofs",        POINTER(__spm_int__)),
                ("layout",      c_int               ),
                ("colptr",      POINTER(__spm_int__)),
                ("rowptr",      POINTER(__spm_int__)),
                ("loc2glob",    POINTER(__spm_int__)),
                ("values",      c_void_p            ),
                ("glob2loc",    POINTER(__spm_int__)),
                ("clustnum",    c_int               ),
                ("clustnbr",    c_int               ),
                ("comm",        pyspm_mpi_comm      ),
                ("replicated",  c_int               ) ]

def pyspm_spmInit( spm ):
    libspm.spmInit.argtypes = [ POINTER(pyspm_spmatrix_t) ]
    libspm.spmInit( spm )

def pyspm_spmInitDist( spm, comm ):
    libspm.spmInitDist.argtypes = [ POINTER(pyspm_spmatrix_t), pyspm_mpi_comm ]
    libspm.spmInitDist( spm, pyspm_convert_comm( comm ) )

def pyspm_spmAlloc( spm ):
    libspm.spmAlloc.argtypes = [ POINTER(pyspm_spmatrix_t) ]
    libspm.spmAlloc( spm )

def pyspm_spmExit( spm ):
    libspm.spmExit.argtypes = [ POINTER(pyspm_spmatrix_t) ]
    libspm.spmExit( spm )

def pyspm_spmCopy( spm_in, spm_out ):
    libspm.spmCopy.argtypes = [ POINTER(pyspm_spmatrix_t),
                                POINTER(pyspm_spmatrix_t) ]
    libspm.spmCopy( spm_in, spm_out )

def pyspm_spmBase( spm, baseval ):
    libspm.spmBase.argtypes = [ POINTER(pyspm_spmatrix_t), c_int ]
    libspm.spmBase( spm, baseval )

def pyspm_spmFindBase( spm ):
    libspm.spmFindBase.argtypes = [ POINTER(pyspm_spmatrix_t) ]
    libspm.spmFindBase.restype = __spm_int__
    return libspm.spmFindBase( spm )

def pyspm_spmGetDegree( spm ):
    libspm.spmGetDegree.argtypes = [ POINTER(pyspm_spmatrix_t) ]
    libspm.spmGetDegree.restype = __spm_int__
    return libspm.spmGetDegree( spm )

def pyspm_spmConvert( ofmttype, ospm ):
    libspm.spmConvert.argtypes = [ c_int, POINTER(pyspm_spmatrix_t) ]
    libspm.spmConvert.restype = c_int
    return libspm.spmConvert( ofmttype, ospm )

def pyspm_spmUpdateComputedFields( spm ):
    libspm.spmUpdateComputedFields.argtypes = [ POINTER(pyspm_spmatrix_t) ]
    libspm.spmUpdateComputedFields( spm )

def pyspm_spmGenFakeValues( spm ):
    libspm.spmGenFakeValues.argtypes = [ POINTER(pyspm_spmatrix_t) ]
    libspm.spmGenFakeValues( spm )

def pyspm_spmScatter( spm_scattered, root, opt_spm_gathered, opt_n,
                      opt_loc2glob, opt_distByColumn, opt_comm ):
    libspm.spmScatter.argtypes = [ POINTER(pyspm_spmatrix_t), c_int,
                                   POINTER(pyspm_spmatrix_t), __spm_int__,
                                   POINTER(__spm_int__), c_int, pyspm_mpi_comm ]
    libspm.spmScatter.restype = c_int
    return libspm.spmScatter( spm_scattered, root, opt_spm_gathered, opt_n,
                              opt_loc2glob.ctypes.data_as( POINTER(__spm_int__) ),
                              opt_distByColumn, pyspm_convert_comm( opt_comm ) )

def pyspm_spmGather( spm_scattered, root, opt_spm_gathered ):
    libspm.spmGather.argtypes = [ POINTER(pyspm_spmatrix_t), c_int,
                                  POINTER(pyspm_spmatrix_t) ]
    libspm.spmGather.restype = c_int
    return libspm.spmGather( spm_scattered, root, opt_spm_gathered )

def pyspm_spmGatherInPlace( spm ):
    libspm.spmGatherInPlace.argtypes = [ POINTER(pyspm_spmatrix_t) ]
    libspm.spmGatherInPlace.restype = c_int
    return libspm.spmGatherInPlace( spm )

def pyspm_spmRedistribute( spm, new_n, newl2g, newspm ):
    libspm.spmRedistribute.argtypes = [ POINTER(pyspm_spmatrix_t), __spm_int__,
                                        POINTER(__spm_int__),
                                        POINTER(pyspm_spmatrix_t) ]
    libspm.spmRedistribute.restype = c_int
    return libspm.spmRedistribute( spm, new_n, newl2g, newspm )

def pyspm_spmNorm( ntype, spm ):
    libspm.spmNorm.argtypes = [ c_int, POINTER(pyspm_spmatrix_t) ]
    libspm.spmNorm.restype = c_double
    return libspm.spmNorm( ntype, spm )

def pyspm_spmNormVec( ntype, spm, x, incx ):
    libspm.spmNormVec.argtypes = [ c_int, POINTER(pyspm_spmatrix_t), c_void_p,
                                   __spm_int__ ]
    libspm.spmNormVec.restype = c_double
    return libspm.spmNormVec( ntype, spm, x, incx )

def pyspm_spmNormMat( ntype, spm, n, A, lda ):
    libspm.spmNormMat.argtypes = [ c_int, POINTER(pyspm_spmatrix_t),
                                   __spm_int__, c_void_p, __spm_int__ ]
    libspm.spmNormMat.restype = c_double
    return libspm.spmNormMat( ntype, spm, n, A, lda )

def pyspm_spmMatVec( trans, alpha, spm, x, beta, y ):
    libspm.spmMatVec.argtypes = [ c_int, c_double, POINTER(pyspm_spmatrix_t),
                                  c_void_p, c_double, c_void_p ]
    libspm.spmMatVec.restype = c_int
    return libspm.spmMatVec( trans, alpha, spm, x, beta, y )

def pyspm_spmMatMat( trans, n, alpha, A, B, ldb, beta, C, ldc ):
    libspm.spmMatMat.argtypes = [ c_int, __spm_int__, c_double,
                                  POINTER(pyspm_spmatrix_t), c_void_p,
                                  __spm_int__, c_double, c_void_p, __spm_int__ ]
    libspm.spmMatMat.restype = c_int
    return libspm.spmMatMat( trans, n, alpha,
                             A.ctypes.data_as( POINTER(pyspm_spmatrix_t) ), B,
                             ldb, beta, C, ldc )

def pyspm_spmScal( alpha, spm ):
    libspm.spmScal.argtypes = [ c_double, POINTER(pyspm_spmatrix_t) ]
    libspm.spmScal( alpha, spm )

def pyspm_spmScalVec( alpha, spm, x, incx ):
    libspm.spmScalVec.argtypes = [ c_double, POINTER(pyspm_spmatrix_t),
                                   c_void_p, __spm_int__ ]
    libspm.spmScalVec( alpha, spm, x, incx )

def pyspm_spmScalMat( alpha, spm, n, A, lda ):
    libspm.spmScalMat.argtypes = [ c_double, POINTER(pyspm_spmatrix_t),
                                   __spm_int__, c_void_p, __spm_int__ ]
    libspm.spmScalMat( alpha, spm, n, A, lda )

def pyspm_spmSort( spm ):
    libspm.spmSort.argtypes = [ POINTER(pyspm_spmatrix_t) ]
    libspm.spmSort.restype = c_int
    return libspm.spmSort( spm )

def pyspm_spmMergeDuplicate( spm ):
    libspm.spmMergeDuplicate.argtypes = [ POINTER(pyspm_spmatrix_t) ]
    libspm.spmMergeDuplicate.restype = __spm_int__
    return libspm.spmMergeDuplicate( spm )

def pyspm_spmSymmetrize( spm ):
    libspm.spmSymmetrize.argtypes = [ POINTER(pyspm_spmatrix_t) ]
    libspm.spmSymmetrize.restype = __spm_int__
    return libspm.spmSymmetrize( spm )

def pyspm_spmCheckAndCorrect( spm_in, spm_out ):
    libspm.spmCheckAndCorrect.argtypes = [ POINTER(pyspm_spmatrix_t),
                                           POINTER(pyspm_spmatrix_t) ]
    libspm.spmCheckAndCorrect.restype = c_int
    return libspm.spmCheckAndCorrect( spm_in, spm_out )

def pyspm_spmGenMat( type, nrhs, spm, alpha, seed, A, lda ):
    libspm.spmGenMat.argtypes = [ c_int, __spm_int__, POINTER(pyspm_spmatrix_t),
                                  c_void_p, c_ulonglong, c_void_p, __spm_int__ ]
    libspm.spmGenMat.restype = c_int
    return libspm.spmGenMat( type, nrhs, spm, alpha, seed, A, lda )

def pyspm_spmGenVec( type, spm, alpha, seed, x, incx ):
    libspm.spmGenVec.argtypes = [ c_int, POINTER(pyspm_spmatrix_t), c_void_p,
                                  c_ulonglong, c_void_p, __spm_int__ ]
    libspm.spmGenVec.restype = c_int
    return libspm.spmGenVec( type, spm, alpha, seed, x, incx )

def pyspm_spmGenRHS( type, nrhs, spm, opt_X, opt_ldx, B, ldb ):
    libspm.spmGenRHS.argtypes = [ c_int, __spm_int__, POINTER(pyspm_spmatrix_t),
                                  c_void_p, __spm_int__, c_void_p, __spm_int__ ]
    libspm.spmGenRHS.restype = c_int
    return libspm.spmGenRHS( type, nrhs, spm, opt_X, opt_ldx, B, ldb )

def pyspm_spmCheckAxb( eps, nrhs, spm, opt_X0, opt_ldx0, B, ldb, X, ldx ):
    libspm.spmCheckAxb.argtypes = [ c_double, __spm_int__,
                                    POINTER(pyspm_spmatrix_t), c_void_p,
                                    __spm_int__, c_void_p, __spm_int__,
                                    c_void_p, __spm_int__ ]
    libspm.spmCheckAxb.restype = c_int
    return libspm.spmCheckAxb( eps, nrhs, spm, opt_X0, opt_ldx0, B, ldb, X, ldx )

def pyspm_spmExtractLocalRHS( nrhs, spm, Bg, ldbg, Bl, ldbl ):
    libspm.spmExtractLocalRHS.argtypes = [ __spm_int__,
                                           POINTER(pyspm_spmatrix_t), c_void_p,
                                           __spm_int__, c_void_p, __spm_int__ ]
    libspm.spmExtractLocalRHS.restype = c_int
    return libspm.spmExtractLocalRHS( nrhs, spm, Bg, ldbg, Bl, ldbl )

def pyspm_spmReduceRHS( nrhs, spm, Bg, ldbg, Bl, ldbl ):
    libspm.spmReduceRHS.argtypes = [ __spm_int__, POINTER(pyspm_spmatrix_t),
                                     c_void_p, __spm_int__, c_void_p,
                                     __spm_int__ ]
    libspm.spmReduceRHS.restype = c_int
    return libspm.spmReduceRHS( nrhs, spm, Bg, ldbg, Bl, ldbl )

def pyspm_spmGatherRHS( nrhs, spm, Bl, ldbl, root, Bg, ldbg ):
    libspm.spmGatherRHS.argtypes = [ __spm_int__, POINTER(pyspm_spmatrix_t),
                                     c_void_p, __spm_int__, c_int, c_void_p,
                                     __spm_int__ ]
    libspm.spmGatherRHS.restype = c_int
    return libspm.spmGatherRHS( nrhs, spm, Bl, ldbl, root, Bg, ldbg )

def pyspm_spmIntConvert( n, input, output ):
    libspm.spmIntConvert.argtypes = [ __spm_int__, c_int_p,
                                      POINTER(__spm_int__) ]
    libspm.spmIntConvert( n, input, output )

def pyspm_spmLoadDist( spm, filename, comm ):
    libspm.spmLoadDist.argtypes = [ POINTER(pyspm_spmatrix_t), c_char_p,
                                    pyspm_mpi_comm ]
    libspm.spmLoadDist.restype = c_int
    return libspm.spmLoadDist( spm, filename, pyspm_convert_comm( comm ) )

def pyspm_spmLoad( spm, filename ):
    libspm.spmLoad.argtypes = [ POINTER(pyspm_spmatrix_t), c_char_p ]
    libspm.spmLoad.restype = c_int
    return libspm.spmLoad( spm, filename )

def pyspm_spmSave( spm, filename ):
    libspm.spmSave.argtypes = [ POINTER(pyspm_spmatrix_t), c_char_p ]
    libspm.spmSave.restype = c_int
    return libspm.spmSave( spm, filename )

def pyspm_spmReadDriver( driver, filename, spm ):
    libspm.spmReadDriver.argtypes = [ c_int, c_char_p,
                                      POINTER(pyspm_spmatrix_t) ]
    libspm.spmReadDriver.restype = c_int
    return libspm.spmReadDriver( driver, filename, spm )

def pyspm_spmReadDriverDist( driver, filename, spm, comm ):
    libspm.spmReadDriverDist.argtypes = [ c_int, c_char_p,
                                          POINTER(pyspm_spmatrix_t),
                                          pyspm_mpi_comm ]
    libspm.spmReadDriverDist.restype = c_int
    return libspm.spmReadDriverDist( driver, filename, spm,
                                     pyspm_convert_comm( comm ) )

def pyspm_spmParseLaplacianInfo( filename, flttype, dim1, dim2, dim3, alpha,
                                 beta, dof ):
    libspm.spmParseLaplacianInfo.argtypes = [ c_char_p, c_int_p,
                                              POINTER(__spm_int__),
                                              POINTER(__spm_int__),
                                              POINTER(__spm_int__),
                                              POINTER(c_double),
                                              POINTER(c_double),
                                              POINTER(__spm_int__) ]
    libspm.spmParseLaplacianInfo.restype = c_int
    return libspm.spmParseLaplacianInfo( filename, flttype, dim1, dim2, dim3,
                                         alpha, beta, dof )

def pyspm_spm2Dense( spm, A ):
    libspm.spm2Dense.argtypes = [ POINTER(pyspm_spmatrix_t), c_void_p ]
    libspm.spm2Dense( spm, A )

def pyspm_spmPrint( spm ):
    libspm.spmPrint.argtypes = [ POINTER(pyspm_spmatrix_t), c_void_p ]
    libspm.spmPrint( spm, None )

def pyspm_spmPrintRHS( spm, nrhs, x, ldx ):
    libspm.spmPrintRHS.argtypes = [ POINTER(pyspm_spmatrix_t), c_int, c_void_p,
                                    __spm_int__, c_void_p ]
    libspm.spmPrintRHS( spm, nrhs, x, ldx, None )

def pyspm_spmPrintInfo( spm ):
    libspm.spmPrintInfo.argtypes = [ POINTER(pyspm_spmatrix_t), c_void_p ]
    libspm.spmPrintInfo( spm, None )

def pyspm_spmExpand( spm_in, spm_out ):
    libspm.spmExpand.argtypes = [ POINTER(pyspm_spmatrix_t),
                                  POINTER(pyspm_spmatrix_t) ]
    libspm.spmExpand( spm_in, spm_out )

def pyspm_spmDofExtend( spm, type, dof, spm_out ):
    libspm.spmDofExtend.argtypes = [ POINTER(pyspm_spmatrix_t), c_int, c_int,
                                     POINTER(pyspm_spmatrix_t) ]
    libspm.spmDofExtend.restype = c_int
    return libspm.spmDofExtend( spm, type, dof, spm_out )

def pyspm_spmBlasGetNumThreads( ):
    libspm.spmBlasGetNumThreads.argtypes = [ ]
    libspm.spmBlasGetNumThreads.restype = c_int
    return libspm.spmBlasGetNumThreads( )

def pyspm_spmBlasSetNumThreads( nt ):
    libspm.spmBlasSetNumThreads.argtypes = [ c_int ]
    libspm.spmBlasSetNumThreads.restype = c_int
    return libspm.spmBlasSetNumThreads( nt )

def pyspm_spmBlasSetNumThreadsOne( ):
    libspm.spmBlasSetNumThreadsOne.argtypes = [ ]
    libspm.spmBlasSetNumThreadsOne.restype = c_int
    return libspm.spmBlasSetNumThreadsOne( )




"""
 @file pastix.py

 PaStiX python wrapper

 @copyright 2017-2024 Bordeaux INP, CNRS (LaBRI UMR 5800), Inria,
                      Univ. Bordeaux. All rights reserved.

 @version 6.4.0
 @author Pierre Ramet
 @author Mathieu Faverge
 @author Louis Poirel
 @date 2024-07-05

"""
from ctypes import *
import numpy as np
import spm
from .enum       import *
from .enum       import __pastix_int__
from .__pastix__ import *

def initParam():
    iparm_ = np.zeros( iparm.size, dtype=__pastix_int__)
    dparm_ = np.zeros( dparm.size, dtype='float64' )
    pypastix_pastixInitParam( iparm_, dparm_ )
    return iparm_, dparm_

def init( iparm, dparm, comm=pypastix_default_comm ):
    pastix_data = c_void_p()
    pypastix_pastixInit( pastix_data, comm, iparm, dparm )
    return pastix_data

def finalize( pastix_data, iparm, dparm ):
    pypastix_pastixFinalize( pastix_data )

def __getnrhs(nrhs, x):
    if nrhs == -1:
        if x.ndim == 1:
            nrhs = 1
        else:
            nrhs = x.shape[1]
    return nrhs

#
# Order structure
#
def orderGet( pastix_data ):
    return pypastix_pastixOrderGet( pastix_data )

def orderGetPermtab( pastix_data, order = None ):
    if not order:
        order = orderGet( pastix_data )

    pyorder = cast( order, POINTER(pypastix_order_t) ).contents

    perm = np.frombuffer( (__pastix_int__ * (pyorder.vertnbr)).from_address( cast(pyorder.permtab, c_void_p).value ), dtype=__pastix_int__ )
    return perm

def orderGetPeritab( pastix_data, order = None ):
    if not order:
        order = orderGet( pastix_data )

    pyorder = cast( order, POINTER(pypastix_order_t) ).contents

    invp = np.frombuffer( (__pastix_int__ * (pyorder.vertnbr)).from_address( cast(pyorder.peritab, c_void_p).value ), dtype=__pastix_int__ )
    return invp

#
# Pastix Tasks
#
def task_analyze( pastix_data, spm ):
    pypastix_pastix_task_analyze( pastix_data, spm.id_ptr )

def task_numfact( pastix_data, spm ):
    pypastix_pastix_task_numfact( pastix_data, spm.id_ptr )

def task_solve( pastix_data, spm, x, nrhs=-1 ):

    m    = x.shape[0]
    nrhs = __getnrhs( nrhs, x )

    x = np.asarray(x, spm.dtype)

    pypastix_pastix_task_solve( pastix_data, m, nrhs, x.ctypes.data_as(c_void_p), m )

def task_refine( pastix_data, spm, b, x, nrhs=-1 ):

    n    = x.shape[0]
    nrhs = __getnrhs( nrhs, x )
    b    = np.asarray( b, spm.dtype )
    ldb  = b.shape[0]
    x    = np.asarray( x, spm.dtype )
    ldx  = x.shape[0]

    if b.dtype != x.dtype:
        raise TypeError( "b and x must use the same arithmetic" )

    pypastix_pastix_task_refine( pastix_data, n, nrhs,
                                 b.ctypes.data_as(c_void_p), ldb,
                                 x.ctypes.data_as(c_void_p), ldx )

#
# Numerical solve subtasks
#
def rhsInit():
    rhs = c_void_p();
    pypastix_pastixRhsInit( rhs )
    return rhs

def rhsFinalize( rhs ):
    pypastix_pastixRhsFinalize( rhs )

def subtask_applyorder( pastix_data, direction, b, Bp ):
    m = b.shape[0]
    if b.ndim == 1:
        n = 1
    else:
        n = b.shape[1]
    ldb = m

    pypastix_pastix_subtask_applyorder( pastix_data, direction, m, n,
                                        b.ctypes.data_as(c_void_p), ldb, Bp )

def subtask_trsm( pastix_data, side, uplo, trans, diag, b ):
    pypastix_pastix_subtask_trsm( pastix_data, side, uplo, trans, diag, b )

def subtask_diag( pastix_data, b ):
    pypastix_pastix_subtask_diag( pastix_data, b )

#
# Schur complement manipulation routines.
#
def setSchurUnknownList( pastix_data, schur_list ):
    n = schur_list.shape[0]
    schur_list = np.array(schur_list, dtype=__pastix_int__)
    pypastix_pastixSetSchurUnknownList( pastix_data, n, schur_list )

def getSchur( pastix_data, S ):
    pypastix_pastixGetSchur( pastix_data,
                             S.ctypes.data_as(c_void_p),
                             S.shape[0] )

def rhsSchurGet( pastix_data, rhsB, B ):
    m = B.shape[0]
    if B.ndim == 1:
        n = 1
    else:
        n = B.shape[1]
    ldb = m
    pypastix_pastixRhsSchurGet( pastix_data, m, n, rhsB,
                                B.ctypes.data_as(c_void_p), ldb )

def rhsSchurSet( pastix_data, B, rhsB ):
    m = B.shape[0]
    if B.ndim == 1:
        n = 1
    else:
        n = B.shape[1]
    ldb = m
    pypastix_pastixRhsSchurSet( pastix_data, m, n,
                                B.ctypes.data_as(c_void_p), ldb, rhsB )




"""

 @file enum.py

 PaStiX python wrapper to define enums and datatypes

 @copyright 2017-2024 Bordeaux INP, CNRS (LaBRI UMR 5800), Inria,
                      Univ. Bordeaux. All rights reserved.

 @version 6.4.0
 @author Pierre Ramet
 @author Mathieu Faverge
 @author Tony Delarue
 @author Selmane Lebdaoui
 @date 2024-06-08

 This file has been automatically generated with gen_wrappers.py

 @ingroup wrap_python

"""
from ctypes import *
import numpy as np

# Start with __ to prevent broadcast to file importing enum
__pastix_int__ = c_int
__pastix_mpi_enabled__ = 1

class iparm:
    verbose                        = 0
    io_strategy                    = 1
    nnzeros                        = 2
    nnzeros_block_local            = 3
    allocated_terms                = 4
    produce_stats                  = 5
    trace                          = 6
    mc64                           = 7
    ordering                       = 8
    ordering_default               = 9
    scotch_mt                      = 10
    scotch_switch_level            = 11
    scotch_cmin                    = 12
    scotch_cmax                    = 13
    scotch_frat                    = 14
    metis_ctype                    = 15
    metis_rtype                    = 16
    metis_no2hop                   = 17
    metis_nseps                    = 18
    metis_niter                    = 19
    metis_ufactor                  = 20
    metis_compress                 = 21
    metis_ccorder                  = 22
    metis_pfactor                  = 23
    metis_seed                     = 24
    metis_dbglvl                   = 25
    amalgamation_lvlblas           = 26
    amalgamation_lvlcblk           = 27
    reordering_split               = 28
    reordering_stop                = 29
    splitting_strategy             = 30
    splitting_levels_projections   = 31
    splitting_levels_kway          = 32
    splitting_projections_depth    = 33
    splitting_projections_distance = 34
    splitting_projections_width    = 35
    min_blocksize                  = 36
    max_blocksize                  = 37
    tasks2d_level                  = 38
    tasks2d_width                  = 39
    allcand                        = 40
    incomplete                     = 41
    level_of_fill                  = 42
    factorization                  = 43
    facto_look_side                = 44
    static_pivoting                = 45
    free_cscuser                   = 46
    schur_fact_mode                = 47
    transpose_solve                = 48
    schur_solv_mode                = 49
    applyperm_ws                   = 50
    refinement                     = 51
    nbiter                         = 52
    itermax                        = 53
    gmres_im                       = 54
    scheduler                      = 55
    thread_nbr                     = 56
    socket_nbr                     = 57
    autosplit_comm                 = 58
    gpu_nbr                        = 59
    gpu_memory_percentage          = 60
    gpu_memory_block_size          = 61
    global_allocation              = 62
    compress_min_width             = 63
    compress_min_height            = 64
    compress_when                  = 65
    compress_method                = 66
    compress_ortho                 = 67
    compress_reltol                = 68
    compress_preselect             = 69
    compress_iluk                  = 70
    mixed                          = 71
    ftz                            = 72
    mpi_thread_level               = 73
    modify_parameter               = 74
    start_task                     = 75
    end_task                       = 76
    float                          = 77
    mtx_type                       = 78
    dof_nbr                        = 79
    size                           = 80

class dparm:
    fill_in            = 0
    epsilon_refinement = 1
    relative_error     = 2
    epsilon_magn_ctrl  = 3
    order_time         = 4
    symbfact_time      = 5
    reorder_time       = 6
    blend_time         = 7
    analyze_time       = 8
    pred_fact_time     = 9
    fact_time          = 10
    fact_flops         = 11
    fact_thflops       = 12
    fact_rlflops       = 13
    fact_energy        = 14
    mem_fr             = 15
    mem_lr             = 16
    solv_time          = 17
    solv_flops         = 18
    solv_thflops       = 19
    solv_rlflops       = 20
    solv_energy        = 21
    refine_time        = 22
    a_norm             = 23
    compress_tolerance = 24
    compress_min_ratio = 25
    size               = 26

class task:
    Init     = 0
    Ordering = 1
    Symbfact = 2
    Analyze  = 3
    Numfact  = 4
    Solve    = 5
    Refine   = 6
    Clean    = 7

class trace:
    Not          = 0
    Numfact      = 16
    Solve        = 32
    FactAndSolve = 48

class verbose:
    Not = 0
    No  = 1
    Yes = 2

class io:
    No        = 0
    Load      = 1
    Save      = 2
    LoadGraph = 4
    SaveGraph = 8
    LoadCSC   = 16
    SaveCSC   = 32

class fact_mode:
    Local = 0
    Schur = 1
    Both  = 2

class solv_mode:
    Local     = 0
    Interface = 1
    Schur     = 2

class refine:
    GMRES    = 0
    CG       = 1
    SR       = 2
    BiCGSTAB = 3

class factotype:
    POTRF = 0
    SYTRF = 1
    GETRF = 2
    PXTRF = 3
    HETRF = 4
    LLH   = 0
    LDLT  = 1
    LU    = 2
    LLT   = 3
    LDLH  = 4

class factolookside:
    FactLeftLooking  = 0
    FactRightLooking = 1

class scheduler:
    Sequential = 0
    Static     = 1
    Parsec     = 2
    StarPU     = 3
    Dynamic    = 4

class ordering:
    Scotch   = 0
    Metis    = 1
    Personal = 2
    PtScotch = 3
    ParMetis = 4

class mpithreadmode:
    MpiNone             = 0
    MpiThreadSingle     = 1
    MpiThreadFunneled   = 2
    MpiThreadSerialized = 3
    MpiThreadMultiple   = 4

class error:
    SUCCESS        = 0
    UNKNOWN        = 1
    ALLOC          = 2
    NOTIMPLEMENTED = 3
    OUTOFMEMORY    = 4
    THREAD         = 5
    INTERNAL       = 6
    BADPARAMETER   = 7
    FILE           = 8
    INTEGER_TYPE   = 9
    IO             = 10
    MPI            = 11

class compress_when:
    Never  = 0
    Begin  = 1
    End    = 2
    During = 3

class compress_method:
    SVD   = 0
    PQRCP = 1
    RQRCP = 2
    TQRCP = 3
    RQRRT = 4
    Nbr   = 5

class compress_ortho:
    OrthoCGS       = 0
    OrthoQR        = 1
    OrthoPartialQR = 2

class split:
    Not             = 0
    Kway            = 1
    KwayProjections = 2

class layout:
    RowMajor = 101
    ColMajor = 102

class trans:
    NoTrans   = 111
    Trans     = 112
    ConjTrans = 113

class uplo:
    Upper      = 121
    Lower      = 122
    UpperLower = 123

class coefside:
    LCoef  = 0
    UCoef  = 1
    LUCoef = 2

class diag:
    NonUnit = 131
    Unit    = 132

class side:
    Left  = 141
    Right = 142

class normtype:
    One       = 171
    Frobenius = 174
    Inf       = 175
    Max       = 177

class dir:
    Forward  = 391
    Backward = 392

class mtxtype:
    General   = trans.NoTrans
    Symmetric = trans.Trans
    Hermitian = trans.ConjTrans
    SymPosDef = trans.ConjTrans + 1
    HerPosDef = trans.ConjTrans + 2




#
# @file __init__.py
#
# SParse Matrix package python module intialization
#
# @copyright 2017-2024 Bordeaux INP, CNRS (LaBRI UMR 5800), Inria,
#                      Univ. Bordeaux. All rights reserved.
#
# @version 1.2.4
# @author Pierre Ramet
# @author Mathieu Faverge
# @author Tony Delarue
# @date 2024-05-29
#
"""
PySpm
=====

Provides
  1. A sparse matrix structure
  2. Mathematical operations over this sparse matrix structure
  3. Driver to read from different file formats and to convert from Scipy package

"""
import ctypes
import ctypes.util

# Load the SPM library
libspm_name = ctypes.util.find_library('spm')
if libspm_name == None:
    raise EnvironmentError("Could not find shared library: spm. "
                           "The path to libspm.so should be in "
                           "$LIBRARY_PATH")

try:
    libspm = ctypes.cdll.LoadLibrary(libspm_name)
except:
    raise EnvironmentError("Could not load shared library: spm. "
                           "The path to libspm.so should be in "
                           "$LD_LIBRARY_PATH or $DYLD_LIBRARY_PATH on MacOS");

from .enum   import *
from .spm    import *

__all__ = [ 'libspm' ]

#__all__.extend(enum.__all__)
#__all__.extend(spm.__all__)



#!/usr/bin/env python3
"""
 @file examples/simple_obj.py

 PaStiX simple python example with the solver object

 @copyright 2017-2024 Bordeaux INP, CNRS (LaBRI UMR 5800), Inria,
                      Univ. Bordeaux. All rights reserved.

 @version 6.4.0
 @author Pierre Ramet
 @author Mathieu Faverge
 @author Louis Poirel
 @date 2024-07-05

"""
##\cond
import pypastix as pastix
import scipy.sparse as sps
import scipy.sparse.linalg
import numpy as np

# Set the problem
n = 1000
A = sps.spdiags([np.ones(n)*i for i in [8., -1., -1., -1., -1.]],
                [0, 1, 3, -1, -3], n, n)
x0 = np.arange(n)

normA = sps.linalg.norm( A )
A = A / normA
b = A.dot(x0)

# Hack to make sure that the mkl is loaded
tmp = np.eye(2).dot(np.ones(2))

# Factorize
solver = pastix.solver(A, verbose=True)

# Solve
x = solver.solve( b )

resid = A.dot( x ) - b
print( "Norm residual: ", np.linalg.norm( resid ) )

# Check the solution
rc = solver.check( x, b, x0, eps=1e-12 )

solver.finalize()

exit(rc)
##\endcond



"""

 @file enum.py

 SPM python wrapper to define enums and datatypes

 @copyright 2017-2024 Bordeaux INP, CNRS (LaBRI UMR 5800), Inria,
                      Univ. Bordeaux. All rights reserved.

 @version 1.2.4
 @author Pierre Ramet
 @author Mathieu Faverge
 @author Tony Delarue
 @date 2024-06-27

 This file has been automatically generated with gen_wrappers.py

 @ingroup wrap_python

"""
from ctypes import *
import numpy as np

# Start with __ to prevent broadcast to file importing enum
__spm_int__ = c_int
__spm_mpi_enabled__ = 1

class verbose:
    Not = 0
    No  = 1
    Yes = 2

class coeftype:
    Pattern   = 0
    Float     = 2
    Double    = 3
    Complex32 = 4
    Complex64 = 5

    @staticmethod
    def getptype ( dtype ):
        np_dict = {
            np.dtype('float32')    : coeftype.Float,
            np.dtype('float64')    : coeftype.Double,
            np.dtype('complex64')  : coeftype.Complex32,
            np.dtype('complex128') : coeftype.Complex64,
        }
        if dtype in np_dict:
            return np_dict[dtype]
        else:
            return -1

    @staticmethod
    def getctype ( flttype ):
        class c_float_complex(Structure):
            _fields_ = [("r",c_float),("i", c_float)]
        class c_double_complex(Structure):
            _fields_ = [("r",c_double),("i", c_double)]

        np_dict = {
            coeftype.Float     : c_float,
            coeftype.Double    : c_double,
            coeftype.Complex32 : c_float_complex,
            coeftype.Complex64 : c_double_complex
        }
        if flttype in np_dict:
            return np_dict[flttype]
        else:
            return -1

    @staticmethod
    def getnptype ( flttype ):
        np_dict = {
            coeftype.Float     : np.dtype('float32'),
            coeftype.Double    : np.dtype('float64'),
            coeftype.Complex32 : np.dtype('complex64'),
            coeftype.Complex64 : np.dtype('complex128')
        }
        if flttype in np_dict:
            return np_dict[flttype]
        else:
            return -1

class fmttype:
    CSC = 0
    CSR = 1
    IJV = 2

class error:
    SUCCESS        = 0
    UNKNOWN        = 1
    ALLOC          = 2
    NOTIMPLEMENTED = 3
    OUTOFMEMORY    = 4
    THREAD         = 5
    INTERNAL       = 6
    BADPARAMETER   = 7
    FILE           = 8
    INTEGER_TYPE   = 9
    IO             = 10
    MPI            = 11

class driver:
    RSA        = 0
    HB         = 1
    IJV        = 2
    MM         = 3
    Laplacian  = 4
    XLaplacian = 5
    Graph      = 6
    SPM        = 7

class rhstype:
    One  = 0
    I    = 1
    RndX = 2
    RndB = 3

class layout:
    RowMajor = 101
    ColMajor = 102

class trans:
    NoTrans   = 111
    Trans     = 112
    ConjTrans = 113

class mtxtype:
    General   = trans.NoTrans
    Symmetric = trans.Trans
    Hermitian = trans.ConjTrans
    SymPosDef = trans.ConjTrans + 1
    HerPosDef = trans.ConjTrans + 2

class uplo:
    Upper      = 121
    Lower      = 122
    UpperLower = 123

class diag:
    NonUnit = 131
    Unit    = 132

class side:
    Left  = 141
    Right = 142

class normtype:
    One       = 171
    Frobenius = 174
    Inf       = 175
    Max       = 177

class dir:
    Forward  = 391
    Backward = 392




#!/usr/bin/env python3
"""
 @file examples/step-by-step.py

 PaStiX step by step python example

 @copyright 2017-2024 Bordeaux INP, CNRS (LaBRI UMR 5800), Inria,
                      Univ. Bordeaux. All rights reserved.

 @version 6.4.0
 @author Pierre Ramet
 @author Mathieu Faverge
 @author Louis Poirel
 @author Vincent Bridonneau
 @date 2024-07-05

"""
##\cond
import pypastix as pastix
import spm
import scipy.sparse as sps
import numpy as np

nrhs = 10

# Hack to make sure that the mkl is loaded
tmp = np.eye(2).dot(np.ones(2))

# Initialize parameters to default values
iparm, dparm = pastix.initParam()

# Startup PaStiX
pastix_data = pastix.init( iparm, dparm )

# Change some parameters
iparm[pastix.iparm.factorization] = pastix.factotype.LU

rc = 0
for nspb in range(1,3):
    # Load a sparse matrix from HB driver
    spmA = spm.spmatrix( None, driver=spm.driver.Laplacian, filename=("%d:%d:%d:4.:1."%(nspb*5, nspb*5, nspb*5)) )
    #spmA = spm.spmatrix( None, driver=driver.HB, filename="$PASTIX_DIR/test/matrix/orsirr.rua" )
    spmA.printInfo()

    # Scale A for low-rank: A / ||A||_f
    norm = spmA.norm()
    spmA.scale( 1. / norm )

    # Perform analyze
    pastix.task_analyze( pastix_data, spmA )

    for nfact in range(1,3):
        # Perform numerical factorization
        pastix.task_numfact( pastix_data, spmA )

        for nsolv in range(1,3):

            # Generate b and x0 vector such that A * x0 = b
            x0, b = spmA.genRHS( spm.rhstype.RndX, nrhs )
            x = b.copy()

            # Perform solve
            pastix.task_solve( pastix_data, spmA, x )

            # Refine the solution
            pastix.task_refine( pastix_data, spmA, b, x )

            # Check solution
            rc |= spmA.checkAxb( x0, b, x )

pastix.finalize( pastix_data, iparm, dparm )

exit(rc)
##\endcond



#!/usr/bin/env python3
"""
 @file examples/schur.py

 PaStiX schur python example

 @copyright 2017-2024 Bordeaux INP, CNRS (LaBRI UMR 5800), Inria,
                      Univ. Bordeaux. All rights reserved.

 @version 6.4.0
 @author Pierre Ramet
 @author Mathieu Faverge
 @author Louis Poirel
 @date 2024-07-05

"""
##\cond
import pypastix as pastix
import spm
import scipy.sparse as sps
import scipy.linalg as la
import numpy as np

# Set matrix A
if True:
    # From Scipy Sparse matrices
    n = 30
    A = sps.spdiags([np.ones(n)*i for i in [4., -1, -1]],
                    [0, 1, -1], n, n)

    x0 = np.arange(n).reshape(n,1)
    # Construct b as b = A * x_0
    b = A.dot(x0)
    spmA = spm.spmatrix( A )

else:
    # Load a sparse matrix from RSA driver
    spmA = spm.spmatrix( None, driver=spm.driver.Laplacian, filename="10:10:10:2.:1." )
    n = spmA.spm_c.gN

    # Generate b and x0 vector such that A * x0 = b
    x0, b = spmA.genRHS( spm.rhstype.RndX )

x = b.copy()

# Hack to make sure that the mkl is loaded
tmp = np.eye(2).dot(np.ones(2))

# Convert the scipy sparse matrix to spm storage format
spmA.printInfo()

# Initialize parameters to default values
iparm, dparm = pastix.initParam()

# Startup PaStiX
pastix_data = pastix.init( iparm, dparm )

factotype = pastix.factotype.LLT
iparm[pastix.iparm.factorization]   = factotype
iparm[pastix.iparm.scheduler]       = pastix.scheduler.Sequential
iparm[pastix.iparm.schur_solv_mode] = pastix.solv_mode.Interface

# Initialize the Schur list as the first third of the elements
nschur = min( int(n / 3.), 5000 )
schurlist = np.arange(nschur) + spmA.findBase()
pastix.setSchurUnknownList ( pastix_data, schurlist )

# Perform analyze
pastix.task_analyze( pastix_data, spmA )

# Perform numerical factorization
pastix.task_numfact( pastix_data, spmA )

# Get the Schur complement
S = np.array( np.zeros( (nschur, nschur) ), order='F', dtype=spmA.dtype )
pastix.getSchur( pastix_data, S )

# Store both sides for linalg
if factotype != pastix.factotype.LU:
    S += la.tril(S, -1).T

schur_x = np.array( np.zeros( nschur ), order='F', dtype=spmA.dtype )

# 1- Apply P to b
Px = pastix.rhsInit()
pastix.subtask_applyorder( pastix_data, pastix.dir.Forward, x, Px )

if factotype == pastix.factotype.LU:
    # 2- Forward solve on the non Schur complement part of the system
    pastix.subtask_trsm( pastix_data, pastix.side.Left, pastix.uplo.Lower, pastix.trans.NoTrans, pastix.diag.Unit, Px )

    pastix.rhsSchurGet( pastix_data, Px, schur_x )

    # 3- Solve the Schur complement part
    schur_x = la.solve(S, schur_x, sym_pos=False)

    pastix.rhsSchurSet( pastix_data, schur_x, Px )

    # 4- Backward solve on the non Schur complement part of the system
    pastix.subtask_trsm( pastix_data, pastix.side.Left, pastix.uplo.Upper, pastix.trans.NoTrans, pastix.diag.NonUnit, Px )

else:
    # 2- Forward solve on the non Schur complement part of the system
    pastix.subtask_trsm( pastix_data, pastix.side.Left, pastix.uplo.Lower, pastix.trans.NoTrans, pastix.diag.NonUnit, Px )

    pastix.rhsSchurGet( pastix_data, Px, schur_x )

    # 3- Solve the Schur complement part
    schur_x = la.solve(S, schur_x, sym_pos=True, lower=True)

    pastix.rhsSchurSet( pastix_data, schur_x, Px )

    # 4- Backward solve on the non Schur complement part of the system
    pastix.subtask_trsm( pastix_data, pastix.side.Left, pastix.uplo.Lower, pastix.trans.ConjTrans, pastix.diag.NonUnit, Px )

# 5- Apply P^t to x
pastix.subtask_applyorder( pastix_data, pastix.dir.Backward, x, Px )
pastix.rhsFinalize( Px )

# Check solution
rc = spmA.checkAxb( x0, b, x )

pastix.finalize( pastix_data, iparm, dparm )

exit(rc)
##\endcond



"""

 @file __pastix__.py

 PaStiX python wrapper

 @copyright 2017-2024 Bordeaux INP, CNRS (LaBRI UMR 5800), Inria,
                      Univ. Bordeaux. All rights reserved.

 @version 6.4.0
 @author Pierre Ramet
 @author Mathieu Faverge
 @author Tony Delarue
 @author Selmane Lebdaoui
 @date 2024-06-08

 This file has been automatically generated with gen_wrappers.py

 @ingroup wrap_python

"""
from ctypes import *
import numpy as np

from . import libpastix
from .enum import __pastix_int__
from .enum import __pastix_mpi_enabled__
from spm import pyspm_spmatrix_t
import spm

if __pastix_mpi_enabled__:
    from mpi4py import MPI
    if MPI._sizeof(MPI.Comm) == sizeof(c_long):
        pypastix_mpi_comm = c_long
    elif MPI._sizeof(MPI.Comm) == sizeof(c_int):
        pypastix_mpi_comm = c_int
    else:
        pypastix_mpi_comm = c_void_p

    pypastix_default_comm = MPI.COMM_WORLD

    def pypastix_convert_comm( comm ):
        comm_ptr = MPI._addressof(comm)
        return pypastix_mpi_comm.from_address(comm_ptr)
else:
    pypastix_mpi_comm = c_int

    pypastix_default_comm = 0

    def pypastix_convert_comm( comm ):
        return c_int(comm)

class pypastix_order_t(Structure):
    _fields_ = [("baseval",      __pastix_int__         ),
                ("vertnbr",      __pastix_int__         ),
                ("cblknbr",      __pastix_int__         ),
                ("permtab",      POINTER(__pastix_int__)),
                ("peritab",      POINTER(__pastix_int__)),
                ("rangtab",      POINTER(__pastix_int__)),
                ("treetab",      POINTER(__pastix_int__)),
                ("selevtx",      POINTER(c_int8)        ),
                ("sndenbr",      __pastix_int__         ),
                ("sndetab",      POINTER(__pastix_int__)),
                ("peritab_exp",  POINTER(__pastix_int__)) ]

def pypastix_pastixOrderInit( ordeptr, baseval, vertnbr, cblknbr, perm, invp,
                              rang, tree ):
    libpastix.pastixOrderInit.argtypes = [ c_void_p, __pastix_int__,
                                           __pastix_int__, __pastix_int__,
                                           POINTER(__pastix_int__),
                                           POINTER(__pastix_int__),
                                           POINTER(__pastix_int__),
                                           POINTER(__pastix_int__) ]
    libpastix.pastixOrderInit.restype = c_int
    return libpastix.pastixOrderInit( ordeptr, baseval, vertnbr, cblknbr,
                                      perm.ctypes.data_as( POINTER(__pastix_int__) ),
                                      invp.ctypes.data_as( POINTER(__pastix_int__) ),
                                      rang.ctypes.data_as( POINTER(__pastix_int__) ),
                                      tree.ctypes.data_as( POINTER(__pastix_int__) ) )

def pypastix_pastixOrderAlloc( ordeptr, vertnbr, cblknbr ):
    libpastix.pastixOrderAlloc.argtypes = [ c_void_p, __pastix_int__,
                                            __pastix_int__ ]
    libpastix.pastixOrderAlloc.restype = c_int
    return libpastix.pastixOrderAlloc( ordeptr, vertnbr, cblknbr )

def pypastix_pastixOrderAllocId( ordeptr, vertnbr ):
    libpastix.pastixOrderAllocId.argtypes = [ c_void_p, __pastix_int__ ]
    libpastix.pastixOrderAllocId.restype = c_int
    return libpastix.pastixOrderAllocId( ordeptr, vertnbr )

def pypastix_pastixOrderExit( ordeptr ):
    libpastix.pastixOrderExit.argtypes = [ c_void_p ]
    libpastix.pastixOrderExit( ordeptr )

def pypastix_pastixOrderBase( ordeptr, baseval ):
    libpastix.pastixOrderBase.argtypes = [ c_void_p, __pastix_int__ ]
    libpastix.pastixOrderBase( ordeptr, baseval )

def pypastix_pastixOrderCheck( ordeptr ):
    libpastix.pastixOrderCheck.argtypes = [ c_void_p ]
    libpastix.pastixOrderCheck.restype = c_int
    return libpastix.pastixOrderCheck( ordeptr )

def pypastix_pastixOrderCopy( ordedst, ordesrc ):
    libpastix.pastixOrderCopy.argtypes = [ c_void_p, c_void_p ]
    libpastix.pastixOrderCopy.restype = c_int
    return libpastix.pastixOrderCopy( ordedst, ordesrc )

def pypastix_pastixOrderGet( pastix_data ):
    libpastix.pastixOrderGet.argtypes = [ c_void_p ]
    libpastix.pastixOrderGet.restype = c_void_p
    return libpastix.pastixOrderGet( pastix_data )

def pypastix_pastixOrderBcast( ordemesh, root, pastix_comm ):
    libpastix.pastixOrderBcast.argtypes = [ c_void_p, c_int, pypastix_mpi_comm ]
    libpastix.pastixOrderBcast( ordemesh, root,
                                pypastix_convert_comm( pastix_comm ) )

def pypastix_pastixOrderGrid( myorder, nx, ny, nz ):
    libpastix.pastixOrderGrid.argtypes = [ c_void_p, __pastix_int__,
                                           __pastix_int__, __pastix_int__ ]
    libpastix.pastixOrderGrid.restype = c_int
    return libpastix.pastixOrderGrid( pointer( myorder ), nx, ny, nz )

def pypastix_pastixOrderLoad( pastix_data, ordeptr ):
    libpastix.pastixOrderLoad.argtypes = [ c_void_p, c_void_p ]
    libpastix.pastixOrderLoad.restype = c_int
    return libpastix.pastixOrderLoad( pastix_data, ordeptr )

def pypastix_pastixOrderSave( pastix_data, ordeptr ):
    libpastix.pastixOrderSave.argtypes = [ c_void_p, c_void_p ]
    libpastix.pastixOrderSave.restype = c_int
    return libpastix.pastixOrderSave( pastix_data, ordeptr )

def pypastix_pastix( pastix_data, pastix_comm, n, colptr, rowptr, values, perm,
                     invp, B, nrhs, iparm, dparm ):
    libpastix.pastix.argtypes = [ c_void_p, pypastix_mpi_comm, __pastix_int__,
                                  POINTER(__pastix_int__),
                                  POINTER(__pastix_int__), c_void_p,
                                  POINTER(__pastix_int__),
                                  POINTER(__pastix_int__), c_void_p,
                                  __pastix_int__, POINTER(__pastix_int__),
                                  POINTER(c_double) ]
    libpastix.pastix.restype = c_int
    return libpastix.pastix( pointer( pastix_data ),
                             pypastix_convert_comm( pastix_comm ), n,
                             colptr.ctypes.data_as( POINTER(__pastix_int__) ),
                             rowptr.ctypes.data_as( POINTER(__pastix_int__) ),
                             values,
                             perm.ctypes.data_as( POINTER(__pastix_int__) ),
                             invp.ctypes.data_as( POINTER(__pastix_int__) ), B,
                             nrhs,
                             iparm.ctypes.data_as( POINTER(__pastix_int__) ),
                             dparm.ctypes.data_as( POINTER(c_double) ) )

def pypastix_pastixInitParam( iparm, dparm ):
    libpastix.pastixInitParam.argtypes = [ POINTER(__pastix_int__),
                                           POINTER(c_double) ]
    libpastix.pastixInitParam( iparm.ctypes.data_as( POINTER(__pastix_int__) ),
                               dparm.ctypes.data_as( POINTER(c_double) ) )

def pypastix_pastixInit( pastix_data, pastix_comm, iparm, dparm ):
    libpastix.pastixInit.argtypes = [ c_void_p, pypastix_mpi_comm,
                                      POINTER(__pastix_int__),
                                      POINTER(c_double) ]
    libpastix.pastixInit( pointer( pastix_data ),
                          pypastix_convert_comm( pastix_comm ),
                          iparm.ctypes.data_as( POINTER(__pastix_int__) ),
                          dparm.ctypes.data_as( POINTER(c_double) ) )

def pypastix_pastixInitWithAffinity( pastix_data, pastix_comm, iparm, dparm,
                                     bindtab ):
    libpastix.pastixInitWithAffinity.argtypes = [ c_void_p, pypastix_mpi_comm,
                                                  POINTER(__pastix_int__),
                                                  POINTER(c_double), c_int_p ]
    libpastix.pastixInitWithAffinity( pointer( pastix_data ),
                                      pypastix_convert_comm( pastix_comm ),
                                      iparm.ctypes.data_as( POINTER(__pastix_int__) ),
                                      dparm.ctypes.data_as( POINTER(c_double) ),
                                      bindtab.ctypes.data_as( c_int_p ) )

def pypastix_pastixFinalize( pastix_data ):
    libpastix.pastixFinalize.argtypes = [ c_void_p ]
    libpastix.pastixFinalize( pointer( pastix_data ) )

def pypastix_pastix_task_analyze( pastix_data, spm ):
    libpastix.pastix_task_analyze.argtypes = [ c_void_p,
                                               POINTER(pyspm_spmatrix_t) ]
    libpastix.pastix_task_analyze.restype = c_int
    return libpastix.pastix_task_analyze( pastix_data, spm )

def pypastix_pastix_task_numfact( pastix_data, spm ):
    libpastix.pastix_task_numfact.argtypes = [ c_void_p,
                                               POINTER(pyspm_spmatrix_t) ]
    libpastix.pastix_task_numfact.restype = c_int
    return libpastix.pastix_task_numfact( pastix_data, spm )

def pypastix_pastix_task_solve( pastix_data, m, nrhs, B, ldb ):
    libpastix.pastix_task_solve.argtypes = [ c_void_p, __pastix_int__,
                                             __pastix_int__, c_void_p,
                                             __pastix_int__ ]
    libpastix.pastix_task_solve.restype = c_int
    return libpastix.pastix_task_solve( pastix_data, m, nrhs, B, ldb )

def pypastix_pastix_task_refine( pastix_data, n, nrhs, B, ldb, X, ldx ):
    libpastix.pastix_task_refine.argtypes = [ c_void_p, __pastix_int__,
                                              __pastix_int__, c_void_p,
                                              __pastix_int__, c_void_p,
                                              __pastix_int__ ]
    libpastix.pastix_task_refine.restype = c_int
    return libpastix.pastix_task_refine( pastix_data, n, nrhs, B, ldb, X, ldx )

def pypastix_pastix_task_solve_and_refine( pastix_data, n, nrhs, B, ldb, X, ldx ):
    libpastix.pastix_task_solve_and_refine.argtypes = [ c_void_p,
                                                        __pastix_int__,
                                                        __pastix_int__,
                                                        c_void_p,
                                                        __pastix_int__,
                                                        c_void_p,
                                                        __pastix_int__ ]
    libpastix.pastix_task_solve_and_refine.restype = c_int
    return libpastix.pastix_task_solve_and_refine( pastix_data, n, nrhs, B, ldb,
                                                   X, ldx )

def pypastix_pastix_subtask_order( pastix_data, spm, myorder ):
    libpastix.pastix_subtask_order.argtypes = [ c_void_p,
                                                POINTER(pyspm_spmatrix_t),
                                                c_void_p ]
    libpastix.pastix_subtask_order.restype = c_int
    return libpastix.pastix_subtask_order( pastix_data, spm, myorder )

def pypastix_pastix_subtask_symbfact( pastix_data ):
    libpastix.pastix_subtask_symbfact.argtypes = [ c_void_p ]
    libpastix.pastix_subtask_symbfact.restype = c_int
    return libpastix.pastix_subtask_symbfact( pastix_data )

def pypastix_pastix_subtask_reordering( pastix_data ):
    libpastix.pastix_subtask_reordering.argtypes = [ c_void_p ]
    libpastix.pastix_subtask_reordering.restype = c_int
    return libpastix.pastix_subtask_reordering( pastix_data )

def pypastix_pastix_subtask_blend( pastix_data ):
    libpastix.pastix_subtask_blend.argtypes = [ c_void_p ]
    libpastix.pastix_subtask_blend.restype = c_int
    return libpastix.pastix_subtask_blend( pastix_data )

def pypastix_pastix_subtask_spm2bcsc( pastix_data, spm ):
    libpastix.pastix_subtask_spm2bcsc.argtypes = [ c_void_p,
                                                   POINTER(pyspm_spmatrix_t) ]
    libpastix.pastix_subtask_spm2bcsc.restype = c_int
    return libpastix.pastix_subtask_spm2bcsc( pastix_data, spm )

def pypastix_pastix_subtask_bcsc2ctab( pastix_data ):
    libpastix.pastix_subtask_bcsc2ctab.argtypes = [ c_void_p ]
    libpastix.pastix_subtask_bcsc2ctab.restype = c_int
    return libpastix.pastix_subtask_bcsc2ctab( pastix_data )

def pypastix_pastix_subtask_sopalin( pastix_data ):
    libpastix.pastix_subtask_sopalin.argtypes = [ c_void_p ]
    libpastix.pastix_subtask_sopalin.restype = c_int
    return libpastix.pastix_subtask_sopalin( pastix_data )

def pypastix_pastix_subtask_applyorder( pastix_data, dir, m, n, B, ldb, Bp ):
    libpastix.pastix_subtask_applyorder.argtypes = [ c_void_p, c_int,
                                                     __pastix_int__,
                                                     __pastix_int__, c_void_p,
                                                     __pastix_int__, c_void_p ]
    libpastix.pastix_subtask_applyorder.restype = c_int
    return libpastix.pastix_subtask_applyorder( pastix_data, dir, m, n, B, ldb,
                                                Bp )

def pypastix_pastix_subtask_trsm( pastix_data, side, uplo, trans, diag, b ):
    libpastix.pastix_subtask_trsm.argtypes = [ c_void_p, c_int, c_int, c_int,
                                               c_int, c_void_p ]
    libpastix.pastix_subtask_trsm.restype = c_int
    return libpastix.pastix_subtask_trsm( pastix_data, side, uplo, trans, diag,
                                          b )

def pypastix_pastix_subtask_diag( pastix_data, b ):
    libpastix.pastix_subtask_diag.argtypes = [ c_void_p, c_void_p ]
    libpastix.pastix_subtask_diag.restype = c_int
    return libpastix.pastix_subtask_diag( pastix_data, b )

def pypastix_pastix_subtask_solve( pastix_data, b ):
    libpastix.pastix_subtask_solve.argtypes = [ c_void_p, c_void_p ]
    libpastix.pastix_subtask_solve.restype = c_int
    return libpastix.pastix_subtask_solve( pastix_data, b )

def pypastix_pastix_subtask_refine( pastix_data, b, x ):
    libpastix.pastix_subtask_refine.argtypes = [ c_void_p, c_void_p, c_void_p ]
    libpastix.pastix_subtask_refine.restype = c_int
    return libpastix.pastix_subtask_refine( pastix_data, b, x )

def pypastix_pastix_subtask_solve_adv( pastix_data, transA, b ):
    libpastix.pastix_subtask_solve_adv.argtypes = [ c_void_p, c_int, c_void_p ]
    libpastix.pastix_subtask_solve_adv.restype = c_int
    return libpastix.pastix_subtask_solve_adv( pastix_data, transA, b )

def pypastix_pastixIsolateUnknowns( pastix_data, n, list ):
    libpastix.pastixIsolateUnknowns.argtypes = [ c_void_p, __pastix_int__,
                                                 POINTER(__pastix_int__) ]
    libpastix.pastixIsolateUnknowns( pastix_data, n,
                                     list.ctypes.data_as( POINTER(__pastix_int__) ) )

def pypastix_pastixSetSchurUnknownList( pastix_data, n, list ):
    libpastix.pastixSetSchurUnknownList.argtypes = [ c_void_p, __pastix_int__,
                                                     POINTER(__pastix_int__) ]
    libpastix.pastixSetSchurUnknownList( pastix_data, n,
                                         list.ctypes.data_as( POINTER(__pastix_int__) ) )

def pypastix_pastixGetSchur( pastix_data, S, lds ):
    libpastix.pastixGetSchur.argtypes = [ c_void_p, c_void_p, __pastix_int__ ]
    libpastix.pastixGetSchur.restype = c_int
    return libpastix.pastixGetSchur( pastix_data, S, lds )

def pypastix_pastixRhsInit( rhs ):
    libpastix.pastixRhsInit.argtypes = [ POINTER(c_void_p) ]
    libpastix.pastixRhsInit.restype = c_int
    return libpastix.pastixRhsInit( pointer( rhs ) )

def pypastix_pastixRhsFinalize( rhs ):
    libpastix.pastixRhsFinalize.argtypes = [ c_void_p ]
    libpastix.pastixRhsFinalize.restype = c_int
    return libpastix.pastixRhsFinalize( rhs )

def pypastix_pastixRhsDoubletoSingle( dB, sB ):
    libpastix.pastixRhsDoubletoSingle.argtypes = [ c_void_p, c_void_p ]
    libpastix.pastixRhsDoubletoSingle.restype = c_int
    return libpastix.pastixRhsDoubletoSingle( dB, sB )

def pypastix_pastixRhsSingleToDouble( sB, dB ):
    libpastix.pastixRhsSingleToDouble.argtypes = [ c_void_p, c_void_p ]
    libpastix.pastixRhsSingleToDouble.restype = c_int
    return libpastix.pastixRhsSingleToDouble( sB, dB )

def pypastix_pastixRhsSchurGet( pastix_data, m, n, rhsB, B, ldb ):
    libpastix.pastixRhsSchurGet.argtypes = [ c_void_p, __pastix_int__,
                                             __pastix_int__, c_void_p, c_void_p,
                                             __pastix_int__ ]
    libpastix.pastixRhsSchurGet.restype = c_int
    return libpastix.pastixRhsSchurGet( pastix_data, m, n, rhsB, B, ldb )

def pypastix_pastixRhsSchurSet( pastix_data, m, n, B, ldb, rhsB ):
    libpastix.pastixRhsSchurSet.argtypes = [ c_void_p, __pastix_int__,
                                             __pastix_int__, c_void_p,
                                             __pastix_int__, c_void_p ]
    libpastix.pastixRhsSchurSet.restype = c_int
    return libpastix.pastixRhsSchurSet( pastix_data, m, n, B, ldb, rhsB )

def pypastix_pastixExpand( pastix_data, spm ):
    libpastix.pastixExpand.argtypes = [ c_void_p, POINTER(pyspm_spmatrix_t) ]
    libpastix.pastixExpand( pastix_data, spm )

def pypastix_pastixGetDiag( pastix_data, x, incx ):
    libpastix.pastixGetDiag.argtypes = [ c_void_p, c_void_p, __pastix_int__ ]
    libpastix.pastixGetDiag.restype = c_int
    return libpastix.pastixGetDiag( pastix_data, x, incx )

def pypastix_pastixGetOptions( argc, argv, iparm, dparm, check, scatter, driver,
                               filename ):
    libpastix.pastixGetOptions.argtypes = [ c_int, c_char_p,
                                            POINTER(__pastix_int__),
                                            POINTER(c_double), c_int_p, c_int_p,
                                            c_int_p, c_char_p ]
    libpastix.pastixGetOptions( argc, pointer( argv ),
                                iparm.ctypes.data_as( POINTER(__pastix_int__) ),
                                dparm.ctypes.data_as( POINTER(c_double) ),
                                check, scatter, driver, pointer( filename ) )

def pypastix_pastixDumpParam( pastix_data ):
    libpastix.pastixDumpParam.argtypes = [ c_void_p ]
    libpastix.pastixDumpParam( pastix_data )

def pypastix_pastixCheckParam( iparm, dparm ):
    libpastix.pastixCheckParam.argtypes = [ POINTER(__pastix_int__),
                                            POINTER(c_double) ]
    libpastix.pastixCheckParam.restype = c_int
    return libpastix.pastixCheckParam(
                                       iparm.ctypes.data_as( POINTER(__pastix_int__) ),
                                       dparm.ctypes.data_as( POINTER(c_double) ) )

def pypastix_pastixBlasGetNumThreads( ):
    libpastix.pastixBlasGetNumThreads.argtypes = [ ]
    libpastix.pastixBlasGetNumThreads.restype = c_int
    return libpastix.pastixBlasGetNumThreads( )

def pypastix_pastixBlasSetNumThreads( nt ):
    libpastix.pastixBlasSetNumThreads.argtypes = [ c_int ]
    libpastix.pastixBlasSetNumThreads.restype = c_int
    return libpastix.pastixBlasSetNumThreads( nt )

def pypastix_pastixBlasSetNumThreadsOne( ):
    libpastix.pastixBlasSetNumThreadsOne.argtypes = [ ]
    libpastix.pastixBlasSetNumThreadsOne.restype = c_int
    return libpastix.pastixBlasSetNumThreadsOne( )




"""
 @file solver.py

 PaStiX python object oriented wrapper

 @copyright 2017-2024 Bordeaux INP, CNRS (LaBRI UMR 5800), Inria,
                      Univ. Bordeaux. All rights reserved.

 @version 6.4.0
 @author Pierre Ramet
 @author Mathieu Faverge
 @author Louis Poirel
 @date 2024-07-05

"""
from .pastix import *
from .enum import *
import spm
import scipy.linalg as la

class solver(object):

    def __init__(self, A=None, **kwargs):
        self.iparm, self.dparm = initParam()

        # Verbose level
        self.verbose = kwargs.setdefault("verbose", verbose.No)
        if not self.verbose: # 0 or False
            self.iparm[iparm.verbose] = verbose.Not
        elif self.verbose==2:
            self.iparm[iparm.verbose] = verbose.Yes
        else: # 1 or True or anything else, default value
            self.iparm[iparm.verbose] = verbose.No

        # Set default factotype based on Matrix properties
        self.mtxtype = kwargs.setdefault("mtxtype", spm.mtxtype.General)
        if self.mtxtype == spm.mtxtype.HerPosDef:
            self.factotype = factotype.LLH
        elif self.mtxtype == spm.mtxtype.SymPosDef:
            self.factotype = factotype.LLT
        elif self.mtxtype == spm.mtxtype.Hermitian:
            self.factotype = factotype.LDLH
        elif self.mtxtype == spm.mtxtype.Symmetric:
            self.factotype = factotype.LDLT
        else:
            self.factotype = factotype.LU

        # Factorization type
        self.factotype = kwargs.setdefault("factotype", self.factotype)
        self.iparm[iparm.factorization] = self.factotype

        # Threads
        self.thread_nbr = kwargs.setdefault("thread_nbr", "auto")
        if self.thread_nbr == "auto":
            self.thread_nbr = -1;
        self.iparm[iparm.thread_nbr] = self.thread_nbr

        self.pastix_data = init( self.iparm, self.dparm )
        if A is not None:
            self.setup(A)

    def setup(self, A):
        """
        Setup the solver object for the classic case w/o Schur complement
        """
        self.A = A
        self.spmA = spm.spmatrix(A)
        if self.verbose:
            self.spmA.printInfo()
        task_analyze(self.pastix_data, self.spmA)
        task_numfact(self.pastix_data, self.spmA)

    def solve(self, b, refine=True):
        """
        Solve and refine the full problem Ax = b with refinement when no Schur Complement is involved
        """
        x = b.copy()
        task_solve(self.pastix_data, self.spmA, x)
        if refine:
            task_refine(self.pastix_data, self.spmA, b, x)
        return x

    def schur(self, A, schur_list, full_matrix=True):
        """
        Setup the solver object to work in Schur complement mode

        full_matrix: boolean, default True. If full_matrix is False
        and a symmetric factorization is used, only the lower part of
        the Schur matrix is stored in self.S
        """

        self.A    = A
        self.spmA = spm.spmatrix(A)
        if self.verbose:
            self.spmA.printInfo()

        self.schur_list = schur_list + self.spmA.findBase()
        setSchurUnknownList(self.pastix_data, self.schur_list)

        task_analyze(self.pastix_data, self.spmA)
        task_numfact(self.pastix_data, self.spmA)

        nschur = len(schur_list)
        self.S = np.zeros( (nschur, nschur), order='F', dtype=A.dtype )
        getSchur( self.pastix_data, self.S )
        if full_matrix and (self.factotype != factotype.LU):
            self.S += la.tril(self.S, -1).T

    def schur_forward(self, x, solv_mode=solv_mode.Interface):
        """
        Solves the forward step of the Schur problem:
               A x = b with A = L S L^h

        This step solves  L f = b with f = S L^h x
        """
        self.iparm[iparm.schur_solv_mode] = solv_mode

        # 1- Apply P to b
        xp = rhsInit()
        subtask_applyorder( self.pastix_data, dir.Forward, x, xp )

        # 2- Forward solve on the non Schur complement part of the system
        if self.factotype in (factotype.LLT, factotype.LLH):
            subtask_trsm( self.pastix_data, side.Left,
                          uplo.Lower, trans.NoTrans,
                          diag.NonUnit, xp )
        else:
            subtask_trsm( self.pastix_data, side.Left,
                          uplo.Lower, trans.NoTrans,
                          diag.Unit, xp )

        # 3- Apply the diagonal step if LDL factorization applied
        if self.factotype in (factotype.LDLT, factotype.LDLH):
            subtask_diag( self.pastix_data, xp )

        self.x  = x
        self.xp = xp

        if x.ndim == 1:
            nrhs = 1
        else:
            nrhs = x.shape[1]

        nschur = len(self.schur_list)
        schurx = np.array( np.zeros( (nschur, nrhs) ), order='F', dtype=self.A.dtype )

        rhsSchurGet( self.pastix_data, xp, schurx )

        return schurx

    def schur_backward(self, y, solv_mode=solv_mode.Interface, refine=True):
        """
        Solves the backward step of the Schur problem:
               A x = b with A = L S L^h

        This step solves L^h x = y with y the solution of L S y = b
        """
        self.iparm[iparm.schur_solv_mode] = solv_mode

        x  = self.x
        xp = self.xp

        rhsSchurSet( self.pastix_data, y, xp )

        # 4- Backward solve on the non Schur complement part of the system
        if self.factotype == factotype.LDLH:
            subtask_trsm( self.pastix_data, side.Left,
                          uplo.Lower, trans.ConjTrans,
                          diag.Unit, xp )
        elif self.factotype == factotype.LDLT:
            subtask_trsm( self.pastix_data, side.Left,
                          uplo.Lower, trans.Trans,
                          diag.Unit, xp )
        elif self.factotype == factotype.LLH:
            subtask_trsm( self.pastix_data, side.Left,
                          uplo.Lower, trans.ConjTrans,
                          diag.NonUnit, xp )
        elif self.factotype == factotype.LLT:
            subtask_trsm( self.pastix_data, side.Left,
                          uplo.Lower, trans.Trans,
                          diag.NonUnit, xp )
        else: # LU
            subtask_trsm( self.pastix_data, side.Left,
                          uplo.Upper, trans.NoTrans,
                          diag.NonUnit, xp )

        #  5- Apply P^t to x
        subtask_applyorder( self.pastix_data, dir.Backward, x, xp )
        rhsFinalize( xp )
        self.xp = None

        return x

    def check( self, x, b, x0=None, nrhs=-1, **kwargs ):
        eps = kwargs.setdefault( 'eps', -1. )
        return self.spmA.checkAxb( x0, b, x, eps=eps )

    def finalize(self):
        finalize(self.pastix_data, self.iparm, self.dparm)
